﻿/**
サブクラス化。

特定のお約束コードが必要。

Examples:
* ----------------------------------------
* class SubClassPush: Push, ISubClass {
* 	private static extern(Windows) LRESULT SubClass_Proc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam) {
		alias EVENTS!(typeof(this)) EVENTLOOP;
* 		
* 		mixin(EVENTLOOP.WM_TOP());
* 		switch(Message) {
* 			case WM_DESTROY:
* 				mixin(EVENTLOOP.WM_DESTROY());
* 			default:
* 				mixin(EVENTLOOP.WM_LAST());
* 		}
* 		assert(false);
* 	}
* 	mixin(SMixInSubClass);
* 	
* 	this(NeGui Owner, ITEM_ID Id) {
* 		super(Owner, Id);
* 		SubClassOnNeGuiConstructor();
* 	}
* 	
* 	protected override {
* 		void OnDestroy() {
* 			SubClassOnNeGuiDestructor();
* 			super.OnDestroy();
* 		}
* 	}
* }
* ----------------------------------------

*/
module etc.negui.event.subclass;

debug(subclass) import std.stdio: wl = writefln;
debug(subclass) void main() {}

import win32.windows;

//import nemuxi.base;
import etc.negui.system.base;
import etc.negui.system.controller;
import etc.negui.negui;


/**
サブクラス化情報保持クラス。

このクラスをそのまま使うんじゃなくてISubClassとSMixInSubClassを組み合わせて使用。

*/
class SubClass: Raii, IController {
	private {
		WNDPROC SrcProc;
		WNDPROC SubProc;
		NeGui gui;
	}

	/**
	*/
	const WNDPROC subProc() {
		return cast(WNDPROC)SubProc;
	}

	/**
	*/
	override protected void Kill() {
		if(SrcProc) {
			if(!gui.setItemInfo(NeGui.GWL.WNDPROC, cast(LRESULT)SrcProc)) {
				throw new NeGuiException(Text("サブクラス化解除失敗"));
			}
			
			SrcProc = null;
		}
	}
	
	/**
	*/
	this(NeGui gui, WNDPROC SubProc)
	in {
		assert(gui);
		assert(SubProc);
	}
	body {
		this.gui     = gui;
		this.SubProc = SubProc;
		super(true);
	}

	/**
	*/
	override bool start() {
		SrcProc = gui.proc(SubProc);
		return SrcProc
			? true
			: false
		;
	}
	/**
	*/
	override bool stop() {
		return super.kill();
	}
	/**
	*/
	override const bool isRunning() {
		return SubProc
			? true
			: false
		;
	}

	/**
	*/
	LRESULT call(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam) {
		assert(SrcProc);
		return CallWindowProc(SrcProc, hWnd, Message, wParam, lParam);
	}
}

/**
サブクラス化を実装するNeGuiが持っとくべきもの。

サブクラス化のサブクラス化に対処するためprivateで実装。
非privateのときにstack overflowに。

privateでinterfaceって規格的にどうなんだろう。
*/
interface ISubClass {
	private LRESULT call(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam);
	
	private void SubClassOnNeGuiConstructor();
	private void SubClassOnNeGuiDestructor();

}
/**
History:
	$(HIST_VER 1.100,
		$(HIST_CHG 名前, TSubClass, SMixInSubClass)
	)
*/
pure nothrow string SMixInSubClass(in string SubclassObject=q{SubclassObject}, in string SubClass_Proc=q{SubClass_Proc}) {
	return `

	private SubClass ` ~ SubclassObject ~ `;
	mixin MixInSubClass!(` ~ SubclassObject ~ `, ` ~ SubClass_Proc ~ `);
	
	`;
}
private mixin template MixInSubClass(alias SubclassObjectName, alias SubClass_ProcName) {
	private override LRESULT call(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam) {
		assert(SubclassObjectName);
		return SubclassObjectName.call(hWnd, Message, wParam, lParam);
	}

	/**
	History:
		$(HIST_VER 1.110,
			$(HIST_P OnCreateからOnSubClassへ。)
		)
	*/
	private override void SubClassOnNeGuiConstructor() {
		SubclassObjectName = new SubClass(this, &SubClass_ProcName);
		SubclassObjectName.start();

		this.OnSubClass();
	}
	private override void SubClassOnNeGuiDestructor() {
		SubclassObjectName.kill;
		delete SubclassObjectName;
	}
}

