﻿/**
サブクラス化。

ボタンのサブクラス化。
*/

import etc.negui.allimport;


// メイン
int NeGuiMain(HINSTANCE hInstance, CommandLine Argument, SHOW Show) {
	auto mywnd=new MyWindow;
	return mywnd.execute;
}
mixin WinMain!(NeGuiMain);

// 独自メッセージ
enum MY_MSG: MESSAGETYPE {
	RDOWN = WM_APP+1,
	MDOWN,
}

// メインウィンドウ
class MyWindow: MainWindow {
	protected {
		// コントロールID
		enum CONTROL: ITEM_ID {
			BUTTON=1,
			LABEL,
		}
		// ボタン
		MyButton button;
		// ラベル
		TextLabel label;
	}
	
	override protected {
		// ウィンドウ生成
		void OnCreate() {
			// コントロール生成
			// ボタン
			button = new MyButton(this, CONTROL.BUTTON);
			button.text = Text("ボタン");
			// ラベル
			label  = new TextLabel(this, CONTROL.LABEL);
			// プロパティ設定
			with(label) {
				horizonAlign   = HORIZON_ALIGN.CENTER;
				verticalCenter = true;
				text           = Text.init;
			}

			// 二分割のパネル生成
			auto panel=new Dual(DIRECTION.VERTICAL);
			panel[0] = button; // 上にボタン
			panel[1] = label;  // 下にラベル
			// 上を25%のサイズ
			panel.sizeInfo.percent = 25;
			// パネルをレイアウトマネージャーの基盤パネルに設定
			layoutManager.basePanel = panel;

			// タイトル設定
			text = Text("SubClass Sample");
		}
		// ウィンドウ生成終了
		void OnCreated() {
			// タイトルバーあり
			caption = true;
			// システムメニューあり
			systemMenu = true;
			// サイズ変更用境界線あり
			frameSize = true;
			// サイズの設定
			size(200, 200);
			// デスクトップの中心に移動
			toCenter(CENTER.DESKTOP);
			// 表示
			setShow(SHOW.SHOWNORMAL);
		}

		// コマンドが来た
		bool OnCommand(ITEM_ID Id, MESSAGETYPE MessageType, NeGui Sender) {
			if(Sender is button) {
				// ボタンから来た
				if(MessageType == Button.EVENT.CLICKED) {
					// クリックされたよ
					label.text = Text("ボタンクリック！");
					// 処理はもう終わる
					return true;
				}
			}
			// 知らないコマンドはスーパークラスへ丸投げ
			return super.OnCommand(Id, MessageType, Sender);
		}

		// なんか来た
		bool OnOtherEvent(UINT Message, WPARAM wParam, LPARAM lParam) {
			// 知っているメッセージだけ処理
			switch(Message) {
				case MY_MSG.RDOWN:
					label.text = Text("右クリック！");
					return true;
				case MY_MSG.MDOWN:
					label.text = Text("ホイールクリック！");
					return true;
				default:
					// 何が来たのか知らないのでスーパークラスへ
					return super.OnOtherEvent(Message, wParam, lParam);
			}
		}
		
	}
}

// 独自ボタン
class MyButton: Push, ISubClass {
	// サブクラス用ウィンドウプロシージャ
	private static extern(Windows) LRESULT SubClass_Proc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam) {
		alias EVENTS!(typeof(this)) EVENTLOOP; // 自分用のイベントを生成

		// 必ず最初に行う
		mixin(EVENTLOOP.WM_TOP());
		switch(Message) {
			// クリックを拾うための前準備
			case WM_RBUTTONUP,   WM_MBUTTONUP:
				mixin(EVENTLOOP.WM_BUTTONUP());
			// クリックを拾うための後準備
			case WM_RBUTTONDOWN, WM_MBUTTONDOWN:
				mixin(EVENTLOOP.WM_BUTTONDOWN());
			// サブクラス化における破棄時のお約束
			case WM_DESTROY:
				mixin(EVENTLOOP.WM_DESTROY());
			// 知らないメッセージはswitchを抜けずに処理
			default:
				mixin(EVENTLOOP.WM_LAST());
		}
		// ここに来ることはない
		assert(false);
	}
	// サブクラス化をある程度自動化
	mixin(SMixInSubClass);
	
	this(NeGui Owner, ITEM_ID Id) {
		super(Owner, Id);
		// サブクラス化生成
		SubClassOnNeGuiConstructor();
	}

	protected override {
		// 右クリック
		bool OnMouseRightClick(MOUSE_KEY Keys, int x, int y) {
			// メッセージを親へ伝える
			parent.send(MY_MSG.RDOWN, NONE, NONE);
			return true;
		}
		// ミドルクリック
		bool OnMouseMiddleClick(MOUSE_KEY Keys, int x, int y) {
			// メッセージを親へ伝える
			parent.send(MY_MSG.MDOWN, NONE, NONE);
			return true;
		}
		// 破棄
		void OnDestroy() {
			// サブクラス化を削除
			SubClassOnNeGuiDestructor();
			super.OnDestroy();
		}
	}
}


